import unittest
from dataclasses import dataclass
from unittest.mock import patch

from pynestor.preview_odoo_nestor import (
    EnvironementConfig,
    PreviewUpScript,
    PreviewUtils,
)
from pynestor.pynestor import NestorDescSet, NestorOpt


def _create_all_spec_sources(project_path):
    return NestorDescSet([NestorOpt("sources.branch", "test_branch")])


@dataclass
class MockInstance:
    name: str = "test"
    url: str = "test"
    existing: bool = True
    password: str = "test"
    verbose: bool = False
    filestore = None
    db = None
    spec = {}

    def exist(self):
        return self.existing

    def version(self, values):
        return 15.0

    def install(self, modules: str):
        pass

    def update(self, modules: str):
        pass

    def delete_and_exit_if_failed(self, return_code: int):
        pass

    def wait(self, up: bool = True, postgres: bool = True, timeout=0):
        pass

    def start(self):
        pass

    def direct_call(self, cde=""):
        pass

    def create(self, odoo_version: str = None, values_set: NestorDescSet = None):
        return type(self)()

    def set_memory_worker(self, workers: int = None, memory_hard: int = None, memory_soft: int = None):
        pass

    @staticmethod
    def list():
        pass

    def edit(self, values_set: NestorDescSet = None):
        pass

    def db_restore_from_s3(
        self,
        dump_path: str,
        alt_dump_path: str = None,
        s3_secret: str = None,
        bucket: str = None,
        set_password_to_all: bool = False,
        verbose: bool = False,
    ):
        pass


class TestEnvironment(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.env_dict = {
            "ENABLE_QUEUE_JOB": "True",
            "NESTOR_NAME": "test",
            "CI_PROJECT_DIR": "../src",
            "CI_BUILD_DIR": "./",
            "CI_PROJECT_PATH": "odoo/",
            "GITLAB_TOKEN": "test_token",
            "ODOO_VERSION": "15.0",
            "CI_COMMIT_REF_NAME": "test_mr",
        }

    def prepare_script(self, env_dict):
        config = EnvironementConfig(env_dict)
        config.apply_default()
        script = PreviewUpScript(config)
        return script

    def test_default_env(self):
        env_dict = self.env_dict.copy()
        script = self.prepare_script(env_dict)
        with patch.object(PreviewUtils, "_create_all_spec_sources") as mock_method:
            mock_method.return_value = NestorDescSet([NestorOpt("sources.branch", "test_branch")])
            values = script.get_spec_values(stage="restore")
        script.log_spec(values)
        self.assertTrue(values["options.queueJobs.enabled"].value)
        # values contient des NestorOpt; NestorOpt.value retourne la chaine
        self.assertEqual(values["options.queueJobs.channels"].value, "root:1")

    def test_always_restore(self):
        env_dict = self.env_dict.copy()
        env_dict.update({"ALWAYS_RESTORE": "True"})
        script = self.prepare_script(env_dict)
        script.inst = MockInstance()
        with patch.object(PreviewUtils, "_create_all_spec_sources") as mock_method:
            mock_method.return_value = NestorDescSet([NestorOpt("sources.branch", "test_branch")])
            with patch.object(type(script), "restore_db") as mock_restore_db:
                with patch.object(type(script), "stop") as mock_stop:
                    with patch.object(type(script), "edit_with_values") as mock_edit_with_values:
                        script.run_script()
                        mock_restore_db.assert_called_with()

    def test_always_restore_false(self):
        env_dict = self.env_dict.copy()
        env_dict.update({"ALWAYS_RESTORE": "false"})
        script = self.prepare_script(env_dict)
        script.inst = MockInstance()
        with patch.object(PreviewUtils, "_create_all_spec_sources") as mock_method:
            mock_method.return_value = NestorDescSet([NestorOpt("sources.branch", "test_branch")])
            with patch.object(type(script), "restore_db") as mock_restore_db:
                with patch.object(type(script), "stop") as mock_stop:
                    with patch.object(type(script), "edit_with_values") as mock_edit_with_values:
                        script.run_script()
                        mock_restore_db.assert_not_called()



import base64
import collections.abc
import os
import sys
import time
from datetime import date
from typing import Union

import gitlab
from dateutil.relativedelta import relativedelta
from kubernetes import client, config
from minio import Minio

######################################
# Python wrapping nestor
######################################
from pynestor.pynestor import (
    NestorDescSet,
    NestorGitDesc,
    NestorOpt,
    ScriptNestorInstance,
    Utils,
)

CURRENT_PATH = os.path.expanduser(os.path.abspath(os.path.curdir))
SUPPORTED_S3 = []


class EnvironementConfig:
    def __init__(self, base_env_vars=None):
        base_env_vars = dict(base_env_vars or {}, **os.environ)
        self.NESTOR_NAME = base_env_vars.get("NESTOR_NAME") or base_env_vars.get("CI_COMMIT_REF_SLUG")
        self.NESTOR_NAME_PREFIX = base_env_vars.get("NESTOR_NAME_PREFIX")
        self.ALWAYS_DELETE: bool = base_env_vars.get("ALWAYS_DELETE", str(False)).capitalize() == str(True).capitalize()
        self.ALWAYS_RESTORE: bool = (
            base_env_vars.get("ALWAYS_RESTORE", str(False)).capitalize() == str(True).capitalize()
        )
        self.CI_COMMIT_REF_NAME = base_env_vars.get("CI_COMMIT_REF_NAME")
        self.CI_PROJECT_DIR = base_env_vars.get("CI_PROJECT_DIR")
        self.CI_BUILDS_DIR = base_env_vars.get("CI_BUILDS_DIR")
        self.CI_PROJECT_PATH = base_env_vars.get("CI_PROJECT_PATH")
        self.GITLAB_TOKEN = base_env_vars.get("GITLAB_TOKEN")
        self.ODOO_VERSION = base_env_vars.get("ODOO_VERSION")
        self.S3_SECRET_PREVIEW = base_env_vars.get("S3_SECRET_PREVIEW")
        self.S3_DUMP_SECRET = base_env_vars.get("S3_DUMP_SECRET")
        self.S3_DUMP_BUCKET = base_env_vars.get("S3_DUMP_BUCKET")
        self.VERBOSE = bool(base_env_vars.get("VERBOSE"))
        self.DATE = base_env_vars.get("DATE")
        self.S3_BUCKET = base_env_vars.get("S3_BUCKET")
        self.ENABLE_QUEUE_JOB = base_env_vars.get("ENABLE_QUEUE_JOB")
        self.PROD_INSTANCE_NAME = base_env_vars.get("PROD_INSTANCE_NAME") or base_env_vars.get("CI_PROJECT_NAME")
        self.DB_NAME = base_env_vars.get("DB_NAME")
        self.ODOO_DEPENDS = base_env_vars.get("ODOO_DEPENDS")
        self.DUMP_PATH = base_env_vars.get("PATH_DUMP")
        self.NO_DB_DUMP: bool = base_env_vars.get("NO_DB_DUMP", str(False)).capitalize() == str(True).capitalize()
        self.MODULES_WITHOUT_DEMO: str = base_env_vars.get("MODULES_WITHOUT_DEMO", str(False).capitalize())
        self.NESTOR_CFG_FILENAME = os.environ.get("NESTOR_CFG_FILENAME", ".nestor-cfg.yml")
        self.BASE_MODULE_UPDATE = os.environ.get("BASE_MODULE_UPDATE", "all")
        self.MODULE_TO_INSTALL = os.environ.get("MODULE_TO_INSTALL")

    def apply_default(self):
        self.CI_PROJECT_PATH = self.CI_PROJECT_PATH or CURRENT_PATH
        self.S3_SECRET_PREVIEW = self.S3_SECRET_PREVIEW or "s3-ndp-preview"
        self.S3_DUMP_SECRET = self.S3_DUMP_SECRET or "s3-ndp-scaleway"
        self.S3_DUMP_BUCKET = self.S3_DUMP_BUCKET or "ndp-production-backups"
        self.ALT_DATE = (date.today() + relativedelta(days=-1)).isoformat() if not self.DATE else None
        self.DATE = self.DATE or date.today().isoformat()
        self.S3_BUCKET = "ndp-preview-" + (self.S3_BUCKET or self.NESTOR_NAME)
        self.DB_NAME = self.DB_NAME or self.PROD_INSTANCE_NAME
        self.ODOO_DEPENDS = self.ODOO_DEPENDS or "odoo-addons/common-modules,odoo-addons/community-addons"
        dump_path_fmt = "%(PROD_INSTANCE_NAME)s/journaliere/%(DB_NAME)s_%(DATE)s.dump"
        self.ALT_DUMP_PATH = (
            self.ALT_DATE
            and dump_path_fmt
            % {
                "PROD_INSTANCE_NAME": self.PROD_INSTANCE_NAME,
                "DB_NAME": self.DB_NAME,
                "DATE": self.ALT_DATE,
            }
            or None
        )
        self.DUMP_PATH = (self.DUMP_PATH or dump_path_fmt) % {
            "PROD_INSTANCE_NAME": self.PROD_INSTANCE_NAME,
            "DB_NAME": self.DB_NAME,
            "DATE": self.DATE,
        }
        self.NESTOR_CFG_FILENAME = self.NESTOR_CFG_FILENAME or ".nestor-cfg.yml"
        self.BASE_MODULE_UPDATE = self.BASE_MODULE_UPDATE or "all"
        self.S3_SECRET_PREVIEW = self.S3_SECRET_PREVIEW or "s3-ndp-preview"
        self.S3_DUMP_SECRET = self.S3_DUMP_SECRET or "s3-ndp-scaleway"
        self.S3_DUMP_BUCKET = self.S3_DUMP_BUCKET or "ndp-production-backups"
        self.DATE = self.DATE or date.today().isoformat()

        if self.NESTOR_NAME_PREFIX:
            self.NESTOR_NAME = f"{self.NESTOR_NAME_PREFIX}{self.NESTOR_NAME}"

    def __repr__(self):
        res = "NESTOR_NAME = " + str(self.NESTOR_NAME)
        res += "\nCI_PROJECT_DIR = " + str(self.CI_PROJECT_DIR)
        res += "\nCI_COMMIT_REF_NAME = " + str(self.CI_COMMIT_REF_NAME)
        res += "\nCI_PROJECT_PATH = " + str(self.CI_PROJECT_PATH)
        res += "\nGITLAB_TOKEN = " + str((self.GITLAB_TOKEN and "*********"))
        res += "\nODOO_VERSION = " + str(self.ODOO_VERSION)
        res += "\nS3_SECRET_PREVIEW = " + str(self.S3_SECRET_PREVIEW)
        res += "\nNO_DB_DUMP = " + str(self.NO_DB_DUMP)
        if not self.NO_DB_DUMP:
            res += "\nS3_DUMP_SECRET = " + str(self.S3_DUMP_SECRET)
            res += "\nS3_DUMP_BUCKET = " + str(self.S3_DUMP_BUCKET)
            res += "\nPROD_INSTANCE_NAME = " + str(self.PROD_INSTANCE_NAME)
            res += "\nDUMP_PATH = " + str(self.DUMP_PATH)
        res += "\nVERBOSE = " + str(self.VERBOSE)
        res += "\nDATE = " + str(self.DATE)
        res += "\nS3_BUCKET = " + str(self.S3_BUCKET)
        res += "\nDB_NAME = " + str(self.DB_NAME)
        res += "\nODOO_DEPENDS = " + str(self.ODOO_DEPENDS)
        res += "\nNESTOR_CFG_FILENAME = " + str(self.NESTOR_CFG_FILENAME)
        res += "\nBASE_MODULE_UPDATE = " + str(self.BASE_MODULE_UPDATE)
        return res


class InteractiveConfig(EnvironementConfig):
    def __init__(self):
        super(InteractiveConfig, self).__init__()
        log("Nom du projet git (Sans le 'odoo-addons/' ex: mettre `aef` pour `odoo-addons/aef`) ?")
        log(f"Mettre un '/' si c'est un path absolut (ex /odoo/v15/tms/core') Defaut: {self.CI_PROJECT_PATH}")
        self.CI_PROJECT_PATH = input("=> ") or self.CI_PROJECT_PATH
        DEFAULT_PROD_INSTANCE_NAME = self.PROD_INSTANCE_NAME or self.CI_PROJECT_PATH
        if self.CI_PROJECT_PATH:
            if not self.CI_PROJECT_PATH.startswith("/"):
                self.CI_PROJECT_PATH = "odoo-addons/" + self.CI_PROJECT_PATH
            else:
                self.CI_PROJECT_PATH = self.CI_PROJECT_PATH.removeprefix("/")
        else:
            log("Un projet est obligatoire")
            sys.exit(1)

        log("Version d'Odoo ? default ", self.ODOO_VERSION)
        self.ODOO_VERSION = input("=> ") or self.ODOO_VERSION

        log(f"Nom de la branche ou du tag || default = {self.CI_COMMIT_REF_NAME  or self.ODOO_VERSION}")
        self.CI_COMMIT_REF_NAME = input("=> ") or self.CI_COMMIT_REF_NAME or self.ODOO_VERSION

        log(f"Nom du nestor (Vide pour utliser '{self.NESTOR_NAME or self.CI_COMMIT_REF_NAME}')")
        self.NESTOR_NAME = input("=> ") or self.NESTOR_NAME or self.CI_COMMIT_REF_NAME

        log(f"Nom de l'instance nestor dans le kube de prod || default '{DEFAULT_PROD_INSTANCE_NAME}'")
        self.PROD_INSTANCE_NAME = input("=> ") or DEFAULT_PROD_INSTANCE_NAME

        log(f"Repertoire du projet ou laisser vide pour utiliser '{CURRENT_PATH}' default={self.CI_PROJECT_DIR}")
        self.CI_PROJECT_DIR = input("=> ") or self.CI_PROJECT_DIR or CURRENT_PATH

        log(f"Prefix à utliser pour le nom sinon '{self.NESTOR_NAME}'")
        self.NESTOR_NAME_PREFIX = input("=> ") or ""

        log("Utiliser un dump de prod ? (OUI par défaut)")
        self.NO_DB_DUMP = input("=>").upper() in ["NON", "FALSE"]

        if not self.NO_DB_DUMP:
            log(f"Nom de la base de backup ? (Vide pour le nom de la prod '{self.PROD_INSTANCE_NAME}')")
            self.DB_NAME = input("=> ") or self.PROD_INSTANCE_NAME

            log(f"Date du dump à utiliser (format ISO YYYY-MM-DD) || default = {date.today().isoformat()}")
            self.DATE = input("=> ") or self.DATE

        log(f"Odoo depends || default = {self.ODOO_DEPENDS}")
        self.ODOO_DEPENDS = input("=> ") or self.ODOO_DEPENDS

        log(f"Token Gitlab (Vide pour utiliser '{self.GITLAB_TOKEN}')")
        self.GITLAB_TOKEN = input("=> ") or self.GITLAB_TOKEN


def log(*args):
    print(" ".join([str(arg) for arg in args]), flush=True, file=sys.stderr)


class PreviewUtils:
    def __init__(self, gitlab_api, env_config):
        self.gitlab_api = gitlab_api
        self.env_config = env_config

    def _try_find_suitable_branch(self, project, branch_to_try_names):
        for branch_name in branch_to_try_names:
            try:
                depends_branch = project.branches.get(branch_name).name
                log("Project %s: branch '%s' found" % (project.name, branch_name))
                return depends_branch
            except gitlab.GitlabGetError:
                log("Project %s: branch not found with name %s" % (project.name, branch_name))
                continue
        return project.default_branch

    def create_spec_sources(self, idx, project_path, branch_to_test) -> NestorGitDesc:
        project = self.gitlab_api.projects.get(project_path)
        branch_founded = self._try_find_suitable_branch(project, branch_to_test)
        log("\tProject %s: use branch '%s'" % (project.name, branch_founded))
        return NestorGitDesc(idx, project_path, branch_founded)

    def _create_all_spec_sources(self, main_project_path):
        current_gl_project = self.gitlab_api.projects.get(main_project_path)
        default_branch = Utils.quote_if_needed(str(current_gl_project.default_branch))
        nestor_values = NestorDescSet([NestorOpt("sources.branch", default_branch)])
        spec_src = self.create_spec_sources(
            idx=0,
            project_path=main_project_path,
            branch_to_test=[
                self.env_config.CI_COMMIT_REF_NAME,
                current_gl_project.default_branch,
                self.env_config.ODOO_VERSION,
            ],
        )
        nestor_values.add(spec_src)
        idx = 2
        for depend in self.env_config.ODOO_DEPENDS.split(","):
            if depend:
                nestor_values.add(
                    self.create_spec_sources(
                        idx=idx,
                        project_path=depend,
                        branch_to_test=[
                            self.env_config.CI_COMMIT_REF_NAME,
                            current_gl_project.default_branch,
                            self.env_config.ODOO_VERSION,
                        ],
                    )
                )
                idx += 1
        return nestor_values

    def get_spec_values(self, sections: collections.abc.Iterable = ("default",), stage="runtime") -> NestorDescSet:
        """Retrieve default value from pynestor configuration file and project's configuration file"""
        cfg_file_name = ".nestor-cfg.yml"
        preview_confs_path = os.path.join(os.path.dirname(__file__), "default-preview-nestor-cfg.yml")
        nestor_cfg_path = os.path.join(self.env_config.CI_PROJECT_DIR, cfg_file_name)

        values = NestorDescSet()
        values.add(NestorOpt("version", self.env_config.ODOO_VERSION))
        values.add(self._get_spec_from_path(preview_confs_path, sections, stage))
        values.add(self._get_spec_from_path(nestor_cfg_path, sections, stage))
        return values

    def _get_spec_from_path(
        self, nestor_cfg_path: str, sections: collections.abc.Iterable, stage: str
    ) -> NestorDescSet:
        values = NestorDescSet()
        log("")
        log("Use cfg file %s" % nestor_cfg_path)
        if os.path.exists(nestor_cfg_path):
            for section in sections:
                values.add(NestorDescSet.parse_nestor_cfg(nestor_cfg_path, section))
                values.add(NestorDescSet.parse_nestor_cfg(nestor_cfg_path, section=section + "-" + stage))
        else:
            log("Cfg file %s not exist" % nestor_cfg_path)
        return values


class ScriptNestor:
    def __init__(self, config: EnvironementConfig):
        self.config = config
        self.inst = ScriptNestorInstance(config.NESTOR_NAME)
        self.gitlab_api = gitlab.Gitlab("https://gitlab.ndp-systemes.fr", config.GITLAB_TOKEN, api_version="4")
        self.preview_utils = PreviewUtils(self.gitlab_api, self.config)

    def delete(self):
        self.inst.delete()
        self.wait_all_down_after_delete()

    def edit_with_values(self, values):
        log("MAJ de l'instance", self.inst.name)
        log(values.to_str(pretty=True))
        log("")
        self.stop()
        self.inst.delete_and_exit_if_failed(self.inst.edit(values))
        time.sleep(10)

    def create_with_values(self, values):
        log("")
        log("Nouvelle instance", self.config.NESTOR_NAME)
        log(values.to_str(pretty=True))
        log("")
        self.inst.create(self.config.ODOO_VERSION, values_set=values)
        self.wait_all_up()

    def wait_all_up(self):
        self.inst.delete_and_exit_if_failed(self.inst.wait(postgres=True, timeout=5 * 60))
        self.inst.delete_and_exit_if_failed(self.inst.wait(timeout=5 * 60))

    def wait_all_down_after_delete(self):
        """Attend que la BDD et odoo soit down sans planter quand la BDD ets déjà supprimée"""
        self.inst.wait(up=False, postgres=True, timeout=5 * 60)
        self.inst.wait(up=False, timeout=5 * 60)

    def stop(self):
        self.inst.delete_and_exit_if_failed(self.inst.stop())
        self.inst.delete_and_exit_if_failed(self.inst.wait(up=False, postgres=False, timeout=5 * 60))  # Timeout de 5min

    def enable_s3_spec(self, values) -> NestorDescSet:
        if self.inst.version(values) not in SUPPORTED_S3:
            log("Version", self.inst.version(values), "don't support S3 in preview: allowed are", SUPPORTED_S3)
            return values
        if values["persistence.s3.secret"]:
            values.add(
                self.inst.filestore.enable_on_s3_spec(
                    s3_secret="s3-ndp-preview", s3_bucket=self.config.S3_BUCKET, values=values
                )
            )
        client = self._get_s3_client()
        if not client.bucket_exists(self.config.S3_BUCKET):
            client.make_bucket(self.config.S3_BUCKET)
        return values

    def enable_queue_job(self, values: NestorDescSet = None) -> NestorDescSet:
        channels = values["options.queueJobs.channels"] or NestorOpt("options.queueJobs.channels", "root:1")
        values.add(channels)
        values.add(NestorOpt("options.queueJobs.enabled", True))
        return values

    def log_spec(self, spec_values: NestorDescSet):
        log("")
        log("Spec Nestor", self.inst.name)
        log(spec_values.to_str(pretty=True))
        log("")

    def run_script(self) -> int:
        log("Nothing to do !!!!!!!!!!")
        return 1

    def _get_s3_client(self):
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            config.load_config()
        v1 = client.CoreV1Api()
        ret = v1.list_namespaced_secret("ndp-test")
        for i in ret.items:
            if (i.metadata.labels or {}).get("type") == "s3":
                if self.config.S3_SECRET_PREVIEW == i.metadata.name:
                    return Minio(
                        endpoint=base64.b64decode(i.data["S3_HOST"]).decode("utf-8"),
                        access_key=base64.b64decode(i.data["S3_ACCESS_KEY"]).decode("utf-8"),
                        secret_key=base64.b64decode(i.data["S3_SECRET_KEY"]).decode("utf-8"),
                        region=base64.b64decode(i.data["S3_REGION"]).decode("utf-8"),
                    )


class PreviewUpScript(ScriptNestor):
    def restore_db(self):
        log("Stop de l'instance pour la restauration")
        self.stop()
        log("Restauration du dump %s" % self.config.DUMP_PATH)
        self.inst.delete_and_exit_if_failed(
            self.inst.db_restore_from_s3(
                self.config.DUMP_PATH,
                self.config.ALT_DUMP_PATH,
                s3_secret=self.config.S3_DUMP_SECRET,
                bucket=self.config.S3_DUMP_BUCKET,
                set_password_to_all=True,
            )
        )

    def get_spec_values(self, *, stage: str) -> NestorDescSet:
        values = self.preview_utils.get_spec_values(
            sections=["default", "preview", self.config.CI_COMMIT_REF_NAME], stage=stage
        )
        values.add(self.preview_utils._create_all_spec_sources(self.config.CI_PROJECT_PATH))
        values = self.enable_s3_spec(values)
        if self.config.ENABLE_QUEUE_JOB:
            values = self.enable_queue_job(values)
        if self.config.NO_DB_DUMP:
            values = self.add_with_demo_option(values)
        if self._is_container(values.get("spec.dockerImage", "")):
            # container image supports only https git fetching
            values = self._change_git_method_to_https(values)
        return values

    def _change_git_method_to_https(self, values: NestorDescSet) -> NestorDescSet:
        """Change the git fetch method to https"""
        for idx in list(range(1, 20)):
            idx = "" if idx == 1 else idx
            opt = "spec.sources.repositories.DEPOT_GIT%s" % idx
            if not values.has_sub_key(opt):
                break
            values.add(
                NestorGitDesc(
                    idx,
                    values.get(opt + ".path").value,
                    values.get(opt + ".branch").value,
                    method="https",
                    secret="ndp-gitlab-pull",
                    server="",
                )
            )
        return values

    def _is_container(self, option: Union[NestorOpt, str]) -> bool:
        """
        cornercut method to identify image of type  registry.ndp-systemes.fr/odoo-cloud/container
        """
        if isinstance(option, NestorOpt):
            option = option.value
        return option.endswith("container")

    def run_script(self):
        if not self.inst.name:
            log("Instance don't have a name")
            return 1
        if self.config.ALWAYS_DELETE:
            if self.inst.exist():
                self.delete()
                # it seems that the instance is not really deleted event if nestor says it is, so we wait a little
                time.sleep(30)
        instance_doesnt_exist = not self.inst.exist()
        if instance_doesnt_exist:
            log("Create instance")
            values = self.get_spec_values(stage="restore")
            self.log_spec(values)
            self.create_with_values(values)

        if self.config.ALWAYS_RESTORE or instance_doesnt_exist:
            if not self.config.NO_DB_DUMP:
                if not instance_doesnt_exist:
                    values = self.get_spec_values(stage="restore")
                    self.edit_with_values(values)
                    self.inst.wait(up=True, postgres=True)
                self.restore_db()
            if self.config.MODULE_TO_INSTALL:
                log("Install -i", self.config.MODULE_TO_INSTALL)
                self.inst.install(self.config.MODULE_TO_INSTALL)

        log("MAJ de l'instance", self.inst.name)
        self.stop()
        values = self.get_spec_values(stage="runtime")
        self.log_spec(values)
        self.edit_with_values(values)
        self.stop()  # On force un stop d'Odoo au cas ou il a restart pour pas bloquer l'étape d'apres
        log("Attente du demarrage postgresql")
        self.inst.delete_and_exit_if_failed(self.inst.wait(up=True, postgres=True))
        time.sleep(10)
        log("Update >> -u", self.config.BASE_MODULE_UPDATE)
        self.inst.delete_and_exit_if_failed(self.inst.update(self.config.BASE_MODULE_UPDATE))
        self.inst.delete_and_exit_if_failed(self.inst.wait(up=True, postgres=True))
        self.inst.delete_and_exit_if_failed(self.inst.start())

        self.wait_all_up()

        log("=======================================================")
        log("Url:", self.inst.url)
        log("Mot de passe:", self.inst.password)
        log("=======================================================")

        return 0

    def add_with_demo_option(self, values):
        return values.merge(NestorDescSet([NestorOpt("options.withoutDemo", self.config.MODULES_WITHOUT_DEMO)]))


class PreviewDownScript(ScriptNestor):
    def run_script(self):
        s3_client = self._get_s3_client()
        self.inst.delete()
        log("Removing S3 bucket", self.config.S3_BUCKET)
        if s3_client.bucket_exists(self.config.S3_BUCKET):
            for obj in s3_client.list_objects(self.config.S3_BUCKET, recursive=True):
                log("delete", self.config.S3_BUCKET, obj.object_name)
                s3_client.remove_object(self.config.S3_BUCKET, obj.object_name)
            s3_client.remove_bucket(self.config.S3_BUCKET)
        log("OK")
        return 0



import enum
import os
import re
import subprocess
import sys
from functools import cached_property
from typing import List, Optional, Set, Union

from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


######################################
# Python wrapping nestor
######################################


class Utils:
    @staticmethod
    def flat_map(f, xs):
        return [y for ys in xs for y in f(ys)]

    @staticmethod
    def quote_if_needed(value):
        if isinstance(value, str):
            try:
                float(value)
                value = "'%s'" % value
            except ValueError:
                pass
        return value

    @staticmethod
    def flat_dico(parent_keys, dico):
        result = []
        for key, value in dico.items():
            new_key = (*parent_keys, key)
            if isinstance(value, dict):
                result += Utils.flat_dico(parent_keys=new_key, dico=value)
            else:
                result.append((new_key, value))
        return result


class NestorDesc:
    def to_set_opt(self) -> "NestorDescSet":
        """Return a list of valid parameters for the --set arg for the 'nestor new' command"""
        raise NotImplementedError()

    def _key_hash(self):
        raise NotImplementedError()

    def __hash__(self):
        return hash(self._key_hash())

    def __eq__(self, other):
        if isinstance(other, NestorDesc):
            return self._key_hash() == other._key_hash()
        return False

    def __str__(self):
        return str(self.to_set_opt())

    def __ge__(self, other):
        if isinstance(other, NestorDesc):
            return self._key_hash().__ge__(other._key_hash())
        raise TypeError

    def __le__(self, other):
        if isinstance(other, NestorDesc):
            return self._key_hash().__le__(other._key_hash())
        raise TypeError

    def __lt__(self, other):
        if isinstance(other, NestorDesc):
            return self._key_hash().__lt__(other._key_hash())
        raise TypeError

    def __gt__(self, other):
        if isinstance(other, NestorDesc):
            return self._key_hash().__gt__(other._key_hash())
        raise TypeError


class NestorOpt(NestorDesc):
    _prefix = "spec."

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def to_set_opt(self):
        """Return a list of valid parameters for the --set arg for the 'nestor new' command"""
        return NestorDescSet([self])

    def _key_hash(self):
        name = self.name
        if not name.startswith(NestorOpt._prefix):
            name = NestorOpt._prefix + name
        return name

    def __str__(self):
        name = self.name
        if not name.startswith(NestorOpt._prefix):
            name = NestorOpt._prefix + name
        return "%s=%s" % (name, Utils.quote_if_needed(self.value))

    def __repr__(self):
        return '%s("%s", %s)' % (
            type(self).__name__,
            self.name,
            Utils.quote_if_needed(self.value),
        )


class NestorDescSet:
    """Représente un ensemble de NestorOpt (qui sont des paires clef / valeur)"""

    def __init__(self, init: List["NestorDesc"] = None):
        self.values: Set[NestorOpt] = set()
        self.add(init or [])

    def add(self, other: Union[List["NestorDescSet"], "NestorDescSet", "NestorDesc", List["NestorDesc"]]) -> None:
        if not isinstance(other, list):
            other = [other]
        for vo in other:
            if isinstance(vo, NestorOpt):
                values = [vo]
            elif isinstance(vo, NestorDesc):
                values = vo.to_set_opt().values
            elif isinstance(vo, NestorDescSet):
                values = vo.values
            else:
                raise ValueError("Can't add type %s" % type(vo))
            for v in values:
                if v in self.values:
                    self.values.remove(v)
                self.values.add(v)

    def merge(self, other: Union["NestorDescSet", List["NestorDescSet"]]):
        if not isinstance(other, list):
            other = [other]
        for vo in other:
            if not isinstance(vo, NestorDescSet):
                raise ValueError("Can't add type %s" % type(vo))
            self.values |= vo.values
        return self

    def __getitem__(self, item: Union[str, "NestorDesc"]) -> Optional[NestorOpt]:
        to_find = item
        if isinstance(to_find, str):
            # __hash__ only on the key not on the value, TODO replace with custom lib internal class
            to_find = NestorOpt(item, 0)

        for el in self.values:
            if el == to_find:
                return el
        return None

    def get(self, item, default=None):
        res = self.__getitem__(item)
        if not res:
            if callable(default):
                return default()
            return default
        return res

    def has_sub_key(self, key: str) -> bool:
        """
        True if the set has a key which is a sub key (for example spec.sources.repositories
        is a subset of spec.sources
        """
        return any(k.name.startswith(key) for k in self.values)

    def __repr__(self):
        return self.values.__repr__()

    def __str__(self):
        return self.values.__str__()

    @staticmethod
    def parse_nestor_cfg(path_nestor_cfg, section):
        nestor_values = NestorDescSet()
        stream = open(path_nestor_cfg)
        data = load(stream, Loader=Loader)
        print("Parsing section [%s] of %s" % (section, path_nestor_cfg))
        for key, value in Utils.flat_dico((), data.get(section, {})):
            try:
                nestor_values.add(NestorOpt(".".join(key), value))
            except TypeError as e:
                print(key, value)
                print(e, flush=True)
                raise e
        return nestor_values

    def to_str(self, pretty=False):
        prefix = pretty and "\n" or ","
        return prefix.join(
            sorted(
                map(
                    lambda it: str(it),
                    self.values,
                )
            )
        )


class NestorGitDesc(NestorDesc):
    """Python class to handle the yml part of the Odoo operator
    branch: ""
          method: ssh
          path: odoo-addons/community-addons
          secret: ""
          server: ""
    """

    _tmpl = "spec.sources.repositories.DEPOT_GIT%(idx)s.%(key)s"

    def __init__(self, idx, path, branch=None, method=None, secret=None, server=None):
        self.idx = idx
        self.path = path
        self.branch = branch
        self.method = method
        self.secret = secret
        self.server = server

    def get_dict_set_opt(self, key):
        return {"idx": self.idx or "", "key": key}

    def to_set_opt(self):
        values = NestorDescSet([NestorOpt(self._tmpl % self.get_dict_set_opt("path"), self.path)])
        if self.branch:
            values.add(
                NestorOpt(
                    self._tmpl % self.get_dict_set_opt("branch"),
                    Utils.quote_if_needed(str(self.branch)),
                )
            )
        if self.method:
            values.add(NestorOpt(self._tmpl % self.get_dict_set_opt("method"), self.method))
        if self.secret:
            values.add(NestorOpt(self._tmpl % self.get_dict_set_opt("secret"), self.secret))
        if self.server:
            values.add(NestorOpt(self._tmpl % self.get_dict_set_opt("server"), self.server))
        return values

    def _key_hash(self):
        return tuple(self.path)


class NestorPersistenceDesc(NestorDesc):
    def __init__(self, s3_secret: str = None, s3_bucket: str = None, enable_s3: bool = True):
        self.s3_secret = s3_secret
        self.s3_bucket = s3_bucket
        self.enable_s3 = enable_s3

    def to_set_opt(self) -> NestorDescSet:
        values = NestorDescSet()
        values.add(NestorOpt("persistence.disabled", self.enable_s3))
        if self.s3_secret:
            values.add(NestorOpt("persistence.s3.secret", self.s3_secret))
        if self.s3_bucket:
            values.add(NestorOpt("persistence.s3.bucket", self.s3_bucket))
        if self.s3_secret and self.s3_bucket:
            values.add(NestorOpt("persistence.s3.allowWrite", True))
            values.add(NestorOpt("persistence.s3.enabled", self.enable_s3))
        return values


class NestorCommand:
    def __init__(self, instance, names, *args, **kwargs):
        if isinstance(names, str):
            names = [names]
        self._n_cmd = ["nestor"] + names + [instance.name]
        self.args = list(args)
        self.kwargs = kwargs

    def call(self) -> int:
        print(self, flush=True)
        return subprocess.call(str(self), shell=True)

    def getoutput(self) -> str:
        print(self, flush=True)
        return subprocess.getoutput(str(self))

    def __str__(self):
        return " ".join([str(i) for i in (self._n_cmd + self.args)])


class NestorInstance:
    def __init__(self, name: str, verbose=False):
        self.name = name
        self.verbose = verbose
        self.filestore = NestorFileStoreCommand(self)
        self.db = NestorDatabaseCmd(self)

    def __repr__(self):
        return "nestor[%s]" % self.name

    def version(self, values: Optional[NestorDescSet] = None) -> int:
        if self.exist():
            value = self.spec["version"]
        elif values:
            value = values["version"] and values["version"].value
        else:
            raise ValueError("Version can't be found")
        return int(float(value))

    def save_spec(self, values: NestorDescSet, path_dir: str):
        full_path = path_dir
        if os.path.splitext(path_dir)[1] != "spec":
            full_path = os.path.join(path_dir, self.name + ".spec")
        with open(os.path.expanduser(full_path), "w+") as spec_file:
            spec_file.write(values.to_str(True))
        return full_path

    @staticmethod
    def new_from_yaml(path_spec, name=None):
        ext = os.path.splitext(path_spec)[1]
        if ext not in ["yaml", "yml"]:
            raise ValueError(f"Le format [{ext}] du fichier {path_spec} n'est pas supporté")
        name = name or os.path.splitext(os.path.basename(path_spec))[0]
        inst = NestorInstance(name)
        inst.direct_call(f"cat {path_spec} | nestor new {name} -")
        return inst

    def exist(self):
        return not bool(self.wait(postgres=True, timeout=2))

    def create(self, odoo_version: str = None, values_set: NestorDescSet = None):
        opt = []
        if odoo_version:
            opt.append("--odoo-version")
            opt.append(str(odoo_version))
        if values_set:
            opt.append("--set")
            opt.append('"%s"' % values_set.to_str())
        self._call("new", *opt)
        return self

    def set_memory_worker(self, workers: int = None, memory_hard: int = None, memory_soft: int = None):
        workers = workers or 2
        memory_soft = memory_soft or 450_000_000  # 450Mo
        memory_hard = memory_hard or 1_000_000_000  # 1000Mo
        self.edit(
            NestorDescSet(
                [
                    NestorOpt("options.limitMemoryHard", memory_hard),
                    NestorOpt("options.limitMemorySoft", memory_soft),
                    NestorOpt("workers", workers),
                ]
            )
        )

    @staticmethod
    def list():
        r = subprocess.getoutput("nestor list -f name")
        return [NestorInstance(s.strip()) for s in r.splitlines()[1:]]

    def edit(self, values_set: NestorDescSet = None) -> int:
        return self._edit(values_set)

    def _edit(self, values_set: NestorDescSet = None) -> int:
        self._clear_cache()
        if self.verbose:
            print(values_set.to_str(pretty=True), flush=True)
        opt = []
        if values_set:
            opt.append("--set")
            opt.append('"%s"' % values_set.to_str())
        return self._call("edit", *opt)

    def filestore_s3_dump(self, s3_secret: str = None, bucket: str = None, verbose: bool = False):
        self._call(["filestore", "dump"], "s3", ".", "--progress", "--bucket", bucket or self.name, "--s3", s3_secret)

    def db_restore_from_s3(
        self,
        dump_path: str,
        alt_dump_path: str = None,
        s3_secret: str = None,
        bucket: str = None,
        set_password_to_all: bool = False,
        verbose: bool = False,
    ) -> int:
        if self.db.restore(
            NestorDatabaseCmd.DBCmdTarget.S3, dump_path, s3_secret, bucket, set_password_to_all, verbose
        ):
            print(f"{dump_path} not found, defaulting to {alt_dump_path}")
            return self.db.restore(
                NestorDatabaseCmd.DBCmdTarget.S3, alt_dump_path, s3_secret, bucket, set_password_to_all, verbose
            )
        return 0

    def wait(self, up: bool = True, postgres: bool = False, timeout: int = None) -> int:
        opt = []
        if up:
            opt.append("up")
        else:
            opt.append("down")
        if postgres:
            opt.append("--postgresql")
        if timeout:
            opt.append("--timeout")
            opt.append(str(timeout))
        return self._call("wait", *opt)

    def stop(self) -> int:
        return self._call("stop")

    def start(self) -> int:
        return self._call("start")

    def restart(self) -> int:
        return self._call("restart")

    def delete(self) -> int:
        print('echo "o" | nestor delete %s' % self.name, flush=True)
        return self.direct_call('echo "o" | nestor delete %(NESTOR_NAME)s')

    def update(self, module_names: Union[str, List[str]]) -> int:
        opt = []
        if isinstance(module_names, str):
            opt.append(module_names)
        else:
            opt.append(",".join(module_names))
        return self._call("update", *opt)

    def install(self, module_names: Union[str, List[str]]) -> int:
        opt = []
        if isinstance(module_names, str):
            opt.append(module_names)
        else:
            opt.append(",".join(module_names))
        return self._call("install", *opt)

    def delete_if_failed(self, result: int) -> bool:
        """delete when the result code is not zero (meaning that previous operation has failed)"""
        print("Delete ?", bool(result), flush=True)
        if result:
            return bool(self.delete())
        return False

    def show(self, output: str = "text"):
        opt = []
        if output:
            opt.append("-o")
            opt.append(output)
        return self._getoutput("show", *opt)

    def _clear_cache(self):
        cache = self.__dict__
        if cache.get("spec"):
            del self.spec
        if cache.get("spec_text"):
            del self.spec_text

    def _call(self, cmd_name: Union[str, List[str]], *opt) -> int:
        return NestorCommand(self, cmd_name, *opt).call()

    def _getoutput(self, cmd_name: str, *opt) -> str:
        return NestorCommand(self, cmd_name, *opt).getoutput()

    def direct_call(self, cmd):
        return subprocess.call(cmd % {"NESTOR_NAME": self.name}, shell=True)

    @cached_property
    def spec(self) -> dict:
        return load(self.show("yaml"), Loader)["spec"]

    @cached_property
    def spec_values(self) -> NestorDescSet:
        desc = NestorDescSet()
        specs = self.show("spec").split(",spec")
        key, value = specs[0].split("=")
        desc.add(NestorOpt(key, value))
        for opt in specs[1:]:
            key, value = opt.split("=")
            desc.add(NestorOpt("spec" + key, value))
        return desc

    @cached_property
    def spec_text(self):
        return self.show("text")

    @property
    def password(self):
        m = re.search(r"Odoo admin \/ DB Manager password: (.*?)\n", self.spec_text)
        if m:
            return m.group(1)
        return None

    @property
    def url(self):
        m = re.search(r"URL: (.*?)\n", self.spec_text)
        if m:
            return m.group(1)
        return None


class NestorDatabaseCmd:
    def __init__(self, inst: NestorInstance):
        self._inst = inst

    class DBCmdTarget(enum.Enum):
        LOCAL = "local"
        S3 = "s3"

    def restore(
        self,
        type: Union[str, DBCmdTarget],
        dump_path: str,
        s3_secret: str = None,
        bucket: str = None,
        set_password_to_all: bool = False,
        verbose: bool = False,
    ) -> int:
        if isinstance(type, NestorDatabaseCmd.DBCmdTarget):
            type = type.value
        opt = [type, dump_path]
        if s3_secret:
            opt.append("--s3")
            opt.append(s3_secret)
        if bucket:
            opt.append("--bucket")
            opt.append(bucket)
        if set_password_to_all:
            opt.append("--set-password-to-all")
        if verbose:
            opt.append("--verbose")
        return NestorCommand(self._inst, ["database", "restore"], *opt).call()

    def dump(
        self,
        type: Union[str, DBCmdTarget],
        dump_path: str,
        s3_secret: str = None,
        bucket: str = None,
        verbose: bool = False,
    ):
        if isinstance(type, NestorDatabaseCmd.DBCmdTarget):
            type = type.value
        opt = [str(type), dump_path]
        if s3_secret:
            opt.append("--s3")
            opt.append(s3_secret)
        if bucket:
            opt.append("--bucket")
            opt.append(bucket)
        if verbose:
            opt.append("--verbose")
        return NestorCommand(self._inst, ["database", "dump"], *opt).call()

    def secret(self):
        raise NotImplementedError

    def port(self):
        raise NotImplementedError


class NestorFileStoreCommand:
    def __init__(self, instance: NestorInstance):
        self.inst = instance

    def enable_on_s3_spec(self, s3_secret: str, s3_bucket: str = None, values: NestorDescSet = None):
        assert s3_secret, "A s3 secret is required"
        assert self.inst.version(values) <= 15, "V15 not supported yet"

        s3_bucket = s3_bucket or self.inst.name
        specs = NestorPersistenceDesc(s3_secret, s3_bucket, enable_s3=True).to_set_opt()
        loads = (values["options.load"] or NestorOpt("options.load", "")).value.split(",")

        branch_name = str(float(self.inst.version(values)))
        if self.inst.version(values) >= 14:
            path = "odoo/cloud-modules"  # New repo only for 14.0 for now
            if "odoo_filestore_s3" not in loads:
                loads.append("odoo_filestore_s3")
            if self.inst.version(values) == 15:
                branch_name = "1.2.3"
        else:
            path = "kubernetes/odoo-modules"
            if "odoo_s3" not in loads:
                loads.append("odoo_s3")
        for idx in list(range(2, 20)):
            opt = "sources.repositories.DEPOT_GIT%s.path" % idx
            if not values[opt]:
                specs.add(NestorGitDesc(idx, path, branch_name).to_set_opt())
                break
        else:
            raise ValueError("Trop de DEPOT_GIT: superieur a 20")
        if self.inst.version(values) <= 12:
            specs.add(NestorOpt("options.load", ",".join(loads)))
        else:
            # les options sont separees par des espaces pour eviter le probleme de --set de nestor qui splitte sur
            # les virgules
            specs.add(NestorOpt("options.load", " ".join(loads)))

        return specs

    def dump_on_s3(self, s3_secret: str, s3_bucket: str = None):
        assert s3_secret, "A s3 secret is required"
        s3_bucket = s3_bucket or self.inst.name

        self.inst._call(
            ["filestore", "dump"], "s3", ".", "--progress", "--bucket", s3_bucket or self.inst.name, "--s3", s3_secret
        )

    def restore_from_s3(self, s3_secret, s3_bucket):
        self.inst._call(
            ["filestore", "restore"],
            "s3",
            ".",
            "--progress",
            "--bucket",
            s3_bucket or self.inst.name,
            "--s3",
            s3_secret,
        )


class ScriptNestorInstance(NestorInstance):
    def __init__(self, name, verbose=False):
        super(ScriptNestorInstance, self).__init__(name, verbose)

    @staticmethod
    def from_inst(nestor_inst: NestorInstance):
        return ScriptNestorInstance(nestor_inst.name, nestor_inst.verbose)

    def delete_and_exit_if_failed(self, result: int):
        self.delete_if_failed(result)
        self.exit_if_failed(result)

    def exit_if_failed(self, result: int):
        if result:
            sys.exit(result)



import argparse
import sys

from pynestor import preview_odoo_nestor
from pynestor.pynestor import NestorInstance


def launch_preview_script(args):
    if args.interactive:
        word = "Démarage" if args.up else "Arret"
        preview_odoo_nestor.log(word + " d'une preview en mode interactif")
        script_config = preview_odoo_nestor.InteractiveConfig()
    else:
        script_config = preview_odoo_nestor.EnvironementConfig()

    script_config.apply_default()
    preview_odoo_nestor.log("====Config====")
    preview_odoo_nestor.log(script_config)
    preview_odoo_nestor.log("==============")

    if args.interactive:
        preview_odoo_nestor.log("Voulez vous continuer avec la configuration si dessus ?")
        rep = input("[y/N] =>")
        if rep.upper() != "y".upper():
            preview_odoo_nestor.log("Exit")
            return 0
    script = preview_odoo_nestor.ScriptNestor(script_config)
    if args.up:
        script = preview_odoo_nestor.PreviewUpScript(script_config)
    if args.down:
        script = preview_odoo_nestor.PreviewDownScript(script_config)
    preview_odoo_nestor.log(args)
    return script.run_script()


def main():
    parsr = argparse.ArgumentParser()
    parsr.add_argument(dest="command", help="Command à utiliser")
    args, other = parsr.parse_known_args()
    print(args, other)
    if args.command == "preview":
        preview_parser = argparse.ArgumentParser()
        preview_parser.add_argument("--down", dest="down", action="store_true")
        preview_parser.add_argument("--up", dest="up", action="store_true")
        preview_parser.add_argument("--interactive", "-i", dest="interactive", action="store_true")
        sys.exit(launch_preview_script(preview_parser.parse_args(other)))
    if args.command == "wait":
        wait_parser = argparse.ArgumentParser()
        wait_parser.add_argument(dest="name", help="Command à utiliser")
        wait_parser.add_argument("--down", dest="down", action="store_false")
        wait_parser.add_argument("--db", dest="db", action="store_true")
        wait_parser.add_argument("--timeout", dest="timeout", type=int, default=None)
        args = wait_parser.parse_args(other)
        print(args)
        inst = NestorInstance(args.name)
        inst.wait(args.down, args.db, timeout=args.timeout)
    if args.command == "exist":
        exist_parser = argparse.ArgumentParser()
        exist_parser.add_argument(dest="name", help="Command à utiliser")
        args = exist_parser.parse_args(other)
        print(args)
        inst = NestorInstance(args.name)
        inst.exist()


if __name__ == "__main__":
    main()
